@@ -9,9 +9,11 @@ import re
 
 from traceback import format_exc
 
-from ..api import DecodeHook
+from . import script_host
+from ..api import decode_if_bytes, walk
 from ..compat import IS_PYTHON3, find_module
 from ..msgpack_rpc import ErrorResponse
+from ..util import format_exc_skip
 
 __all__ = ('Host')
 
@@ -40,25 +42,47 @@ class Host(object):
             'specs': self._on_specs_request,
             'shutdown': self.shutdown
         }
-        self._nvim_encoding = nvim.options['encoding']
-        if IS_PYTHON3 and isinstance(self._nvim_encoding, bytes):
-            self._nvim_encoding = self._nvim_encoding.decode('ascii')
+
+        # Decode per default for Python3
+        self._decode_default = IS_PYTHON3
+
+    def _on_async_err(self, msg):
+        self.nvim.err_write(msg, async=True)
 
     def start(self, plugins):
         """Start listening for msgpack-rpc requests and notifications."""
-        self.nvim.session.run(self._on_request,
-                              self._on_notification,
-                              lambda: self._load(plugins))
+        self.nvim.run_loop(self._on_request,
+                           self._on_notification,
+                           lambda: self._load(plugins),
+                           err_cb=self._on_async_err)
 
     def shutdown(self):
         """Shutdown the host."""
         self._unload()
-        self.nvim.session.stop()
+        self.nvim.stop_loop()
+
+    def _wrap_function(self, fn, sync, decode, nvim_bind, name, *args):
+        if decode:
+            args = walk(decode_if_bytes, args, decode)
+        if nvim_bind is not None:
+            args.insert(0, nvim_bind)
+        try:
+            return fn(*args)
+        except Exception:
+            if sync:
+                msg = ("error caught in request handler '{} {}':\n{}"
+                       .format(name, args, format_exc_skip(1, 5)))
+                raise ErrorResponse(msg)
+            else:
+                msg = ("error caught in async handler '{} {}'\n{}\n"
+                       .format(name, args, format_exc_skip(1, 5)))
+                self._on_async_err(msg + "\n")
+                raise
 
     def _on_request(self, name, args):
         """Handle a msgpack-rpc request."""
-        if IS_PYTHON3 and isinstance(name, bytes):
-            name = name.decode(self._nvim_encoding)
+        if IS_PYTHON3:
+            name = decode_if_bytes(name)
         handler = self._request_handlers.get(name, None)
         if not handler:
             msg = self._missing_handler_error(name, 'request')
@@ -72,23 +96,17 @@ class Host(object):
 
     def _on_notification(self, name, args):
         """Handle a msgpack-rpc notification."""
-        if IS_PYTHON3 and isinstance(name, bytes):
-            name = name.decode(self._nvim_encoding)
+        if IS_PYTHON3:
+            name = decode_if_bytes(name)
         handler = self._notification_handlers.get(name, None)
         if not handler:
             msg = self._missing_handler_error(name, 'notification')
             error(msg)
-            self.nvim.err_write(msg + "\n")
+            self._on_async_err(msg + "\n")
             return
 
         debug('calling notification handler for "%s", args: "%s"', name, args)
-        try:
-            handler(*args)
-        except Exception as err:
-            msg = ("error caught in async handler '{} {}':\n{!r}\n{}"
-                   .format(name, args, err, format_exc(5)))
-            self.nvim.err_write(msg, async=True)
-            raise
+        handler(*args)
 
     def _missing_handler_error(self, name, kind):
         msg = 'no {} handler registered for "{}"'.format(kind, name)
@@ -105,15 +123,14 @@ class Host(object):
             if path in self._loaded:
                 error('{0} is already loaded'.format(path))
                 continue
-            if path == "script_host.py":
-                directory = os.path.dirname(__file__)
-                name = "script_host"
-            else:
-                directory, name = os.path.split(os.path.splitext(path)[0])
-            file, pathname, description = find_module(name, [directory])
-            handlers = []
             try:
-                module = imp.load_module(name, file, pathname, description)
+                if path == "script_host.py":
+                    module = script_host
+                else:
+                    directory, name = os.path.split(os.path.splitext(path)[0])
+                    file, pathname, descr = find_module(name, [directory])
+                    module = imp.load_module(name, file, pathname, descr)
+                handlers = []
                 self._discover_classes(module, handlers, path)
                 self._discover_functions(module, handlers, path)
                 if not handlers:
@@ -151,40 +168,37 @@ class Host(object):
     def _discover_functions(self, obj, handlers, plugin_path):
         def predicate(o):
             return hasattr(o, '_nvim_rpc_method_name')
-        specs = []
-        objenc = getattr(obj, '_nvim_encoding', None)
-        for _, fn in inspect.getmembers(obj, predicate):
-            enc = getattr(fn, '_nvim_encoding', objenc)
-            if fn._nvim_bind:
-                # bind a nvim instance to the handler
-                fn2 = functools.partial(fn, self._configure_nvim_for(fn))
-                # copy _nvim_* attributes from the original function
-                self._copy_attributes(fn, fn2)
-                fn = fn2
-            decodehook = self._decodehook_for(enc)
-            if decodehook is not None:
-                decoder = lambda fn, hook, *args: fn(*hook.walk(args))
-                fn2 = functools.partial(decoder, fn, decodehook)
-                self._copy_attributes(fn, fn2)
-                fn = fn2
 
-            # register in the rpc handler dict
+        specs = []
+        objdecode = getattr(obj, '_nvim_decode', self._decode_default)
+        for _, fn in inspect.getmembers(obj, predicate):
+            sync = fn._nvim_rpc_sync
+            decode = getattr(fn, '_nvim_decode', objdecode)
+            nvim_bind = None
+            if fn._nvim_bind:
+                nvim_bind = self._configure_nvim_for(fn)
+
             method = fn._nvim_rpc_method_name
             if fn._nvim_prefix_plugin_path:
                 method = '{0}:{1}'.format(plugin_path, method)
-            if fn._nvim_rpc_sync:
+
+            fn_wrapped = functools.partial(self._wrap_function, fn,
+                                           sync, decode, nvim_bind, method)
+            self._copy_attributes(fn, fn_wrapped)
+            # register in the rpc handler dict
+            if sync:
                 if method in self._request_handlers:
                     raise Exception(('Request handler for "{0}" is ' +
                                     'already registered').format(method))
-                self._request_handlers[method] = fn
+                self._request_handlers[method] = fn_wrapped
             else:
                 if method in self._notification_handlers:
                     raise Exception(('Notification handler for "{0}" is ' +
                                     'already registered').format(method))
-                self._notification_handlers[method] = fn
+                self._notification_handlers[method] = fn_wrapped
             if hasattr(fn, '_nvim_rpc_spec'):
                 specs.append(fn._nvim_rpc_spec)
-            handlers.append(fn)
+            handlers.append(fn_wrapped)
         if specs:
             self._specs[plugin_path] = specs
 
@@ -195,25 +209,16 @@ class Host(object):
                 setattr(fn2, attr, getattr(fn, attr))
 
     def _on_specs_request(self, path):
-        if IS_PYTHON3 and isinstance(path, bytes):
-            path = path.decode(self._nvim_encoding)
+        if IS_PYTHON3:
+            path = decode_if_bytes(path)
         if path in self._load_errors:
             self.nvim.out_write(self._load_errors[path] + '\n')
         return self._specs.get(path, 0)
 
-    def _decodehook_for(self, encoding):
-        if IS_PYTHON3 and encoding is None:
-            encoding = True
-        if encoding is True:
-            encoding = self._nvim_encoding
-        if encoding:
-            return DecodeHook(encoding)
-
     def _configure_nvim_for(self, obj):
         # Configure a nvim instance for obj (checks encoding configuration)
         nvim = self.nvim
-        encoding = getattr(obj, '_nvim_encoding', None)
-        hook = self._decodehook_for(encoding)
-        if hook is not None:
-            nvim = nvim.with_hook(hook)
+        decode = getattr(obj, '_nvim_decode', self._decode_default)
+        if decode:
+            nvim = nvim.with_decode(decode)
         return nvim
